#!/usr/bin/python
# -*- coding: utf-8 -*-
__author__ = 'Administrator'
"""
Version: 1.0
Created on 2015-04-01
"""

import numpy as np
import os
import cPickle
import sys
import getopt
import time
import scan
from collections import Counter
from Bio import SeqIO
from sklearn.externals import joblib

USAGE = """
USAGE
    spar.py  [-i <input file> [-d <database_dir>] [-o <output_dir>] [-n <num_threads >] [-h] [-v]

OPTIONAL ARGUMENTS
    -i <input_file>   : dataset file containing protein sequences as FASTA file format.
    -d <database_dir> : directory location of database for HMMER3.
    -o <output_dir>   : a directory containing the results of prediction of each sample.
                        Default = "current time".
    -n <num_threads>  : number of threads (CPUs) to use for parallel process; int value.
                        Default = 1.
    -v                : print the detail alignment and interface topology of domain information.

    -h                : print USAGE, DESCRIPTION and ARGUMENTS messages.
"""

VERSION = """
DESCRIPTION
   Name               : SPAR (Self-interacting Protein Analyzer)
   Version            : 1.0
   Update Time        : 2015-04-01
"""
PATH = sys.path[0] + '/../'
CUTOFF90 = 5.620
CUTOFF95 = 5.797
AA = 'ARNDCQEGHILKMFPSTWYV'

GROUP = {'-': '0', '.': '0', '*': '0',
         'A': '1', 'G': '1', 'V': '1',
         'I': '2', 'L': '2', 'F': '2', 'P': '2',
         'Y': '3', 'M': '3', 'T': '3', 'S': '3',
         'H': '4', 'N': '4', 'Q': '4', 'W': '4',
         'R': '5', 'K': '5',
         'D': '6', 'E': '6',
         'C': '7', 'U': '7',
         'X': '8', 'B': '8', 'Z': '8'}

rank = np.loadtxt(PATH + 'data/rank.tab', dtype=np.str)[:195, 1].astype(np.int16)
FILE = open(PATH + 'data/prop.pkl', 'rb')
PROP = cPickle.load(FILE)
FILE = open(PATH + 'data/3did.pkl', 'rb')
DID3 = cPickle.load(FILE)


def _dpc(seq):
    keys0 = [i + j for i in AA for j in AA]
    peps = [str(seq[i:i + 2]) for i in range(len(seq) + 1 - 2)]
    count = Counter(peps)
    sample = [count[key] * 1.0 / len(peps) for key in keys0]
    return sample


def _ct(seq):
    keys = [i + j + k for i in '1234567' for j in '1234567' for k in '1234567']
    codes = ''.join([GROUP[res] for res in seq])
    ctps = [str(codes[i:i + 3]) for i in range(len(seq))]
    count = Counter(ctps)
    sample = [count[key] * 1.0 / len(ctps) for key in keys]
    return sample


def _ac(seq, lag=30):
    keys = ['H', 'VSC', 'P1', 'P2', 'SASA', 'NCISC']
    sample = [0] * (lag * len(keys))
    for k, key in enumerate(keys):
        code = [PROP[res][key] for res in seq]
        mean = np.mean(code)
        for j in range(1, lag + 1):
            if len(seq) > j:
                acvs = [(code[i] - mean) * (code[i + j] - mean) for i in range(len(code) - j)]
                sample[(j - 1) * len(keys) + k] = np.sum(acvs) / (len(seq) - j)
            else:
                sample[(j - 1) * len(keys) + k] = 0
    return sample


def _mac(seq, lag=30):
    keys = ['H', 'VSC', 'P1', 'P2', 'SASA', 'NCISC']
    sample = [0] * (lag * len(keys))
    for k, key in enumerate(keys):
        code = [PROP[res][key] for res in seq]
        mean = np.mean(code)
        std = np.std(code)
        for j in range(1, lag + 1):
            if len(code) > j:
                acvs = [(code[i] - mean) * (code[i + j] - mean) for i in range(len(code) - j)]
                sample[(j - 1) * len(keys) + k] = np.sum(acvs) / (std * (len(code) - j))
            else:
                sample[(j - 1) * len(keys) + k] = 0
    return sample


def find(ori, substr, i):
    count = 0
    while i > 0:
        index = ori.find(substr)
        if index == -1:
            return -1
        else:
            ori = ori[index + 1:]
            i -= 1
            count += index + 1
    return count - 1


def _ld(seq):
    keys0 = [i for i in '1234567']
    keys1 = ['%s%s' % (i, j) for i in keys0 for j in keys0 if i < j]
    codes = ''.join([GROUP[res] for res in seq])
    div25, div50, div75, div = len(codes) / 4, len(codes) / 2, len(codes) * 3 / 4, len(codes)
    regions = [
        codes[:div25], codes[div25:div50], codes[div50:div75], codes[div75:],
        codes[:div50], codes[div50:],
        codes[div25:], codes[:div75],
        codes[div25:div75], codes[(div / 8):(div * 7 / 8)]
    ]
    sample = []
    for region in regions:
        count0 = Counter(region)
        C = [count0[key] * 1.0 / len(region) for key in keys0]
        data = [''.join(sorted(region[i:i + 2])) for i in range(len(region) - 1) if region[i] != region[i + 1]]
        count1 = Counter(data)
        T = [count1[key] * 1.0 / len(data) for key in keys1] if len(data) != 0 else [0] * len(keys1)
        D = []
        for i in keys0:
            span25, span50, span75, span100 = count0[i] / 4, count0[i] / 2, count0[i] * 3 / 4, count0[i]
            if i in region:
                d0 = region.index(i)
                d25 = find(region, i, span25)
                d50 = find(region, i, span50)
                d75 = find(region, i, span75)
                d100 = find(region, i, span100)
                D += [d0, d25, d50, d75, d100]
            else:
                D += [0, 0, 0, 0, 0]
        sample += C + T + D
    return sample


def interface(hsps):
    result = '=== The Information of Homodomain Interaction Information in 3did Database ===\n'
    for hsp in hsps:
        if hsp.hit_id in DID3 and hsp.hit_id in DID3[hsp.hit_id]:
            start = hsp.hit_start
            spar = '#DESCRIBE  %s\n' % hsp.describe + \
                   '#LENGTH    %d\n' % hsp.length + \
                   '#HMM_ACC   %s\n' % hsp.acc + \
                   '#HMM_NAME  %s\n' % hsp.hit_id + \
                   '#HMM_START %d\n' % (hsp.hit_start + 1) + \
                   '#HMM_END   %d\n' % hsp.hit_end + \
                   '#SEQ_START %d\n' % (hsp.query_start + 1) + \
                   '#SEQ_END   %d\n' % hsp.query_end + \
                   '#E-VALUE   %.1e\n' % hsp.evalue_cond
            hmms = hsp.hit.seq.__str__()
            inf = ''
            for key, value in DID3[hsp.hit_id][hsp.hit_id].items():
                inf += '#' + key + ' ' * 6
                gap = 0
                for i, hmm in enumerate(hmms):
                    if hmm == '-' or hmm == '.':
                        gap += 1
                        inf += '-'
                    elif i + start - gap in value:
                        inf += '+'
                    else:
                        inf += '-'
                inf += '\n'
            result += '%s#HMM       %s\n#SEQ       %s\n%s' % (spar, hmms, hsp.query.seq.__str__(), inf)
    return result


def substitution(hsps):
    homos = []
    for hsp in hsps:
        homo_topo = set()
        if hsp.hit_id in DID3 and hsp.hit_id in DID3[hsp.hit_id]:
            for topo in DID3[hsp.hit_id][hsp.hit_id].values():
                homo_topo.update(topo)
        pos = hsp.hit_start
        last = -1
        gap = 0
        hit = hsp.hit.seq.__str__().upper().replace('.', '-')
        seq = hsp.query.seq.__str__().upper().replace('.', '-')
        for i in range(len(hit)):
            if hit[i] is '-':
                gap += 1
            else:
                if pos in homo_topo:
                    if gap != 0:
                        if last + gap + 1 == i:
                            homos += ['-' + res for res in seq[i + 1:i + gap + 1]]
                        gap = 0
                    homos.append(hit[i] + seq[i])
                    last = i
                pos += 1
        homos += ['--'] * (len(homo_topo) - len(homos))
    return homos


def _rsi(hsps):
    keys = [i + j for i in AA + '-' for j in AA + '-']
    HOMO = substitution(hsps)
    nhomo = Counter(HOMO)
    sample = [nhomo[i] * 1.0 / len(HOMO) for i in keys] if len(HOMO) != 0 else [0] * len(keys)
    return sample


def encode(fasta, db, verbose=False):
    records = SeqIO.parse(fasta, 'fasta')
    keys = []
    samples = []
    infs = []
    timeArray = time.localtime(int(time.time()))
    temp = time.strftime("%Y%m%d%H%M%S", timeArray) + '.tmp'
    temp = temp if os.path.isabs(fasta) else PATH + temp
    for record in records:
        head = record.id
        seq = str(record.seq).upper()
        dpc = _dpc(seq)
        ct = _ct(seq)
        ac = _ac(seq)
        mac = _mac(seq)
        ld = _ld(seq)
        SeqIO.write(record, temp, 'fasta')
        try:
            hmmscan = scan.Scan(db, temp)
            hmmscan.search()
            rsi = _rsi(hmmscan.records[0])
            keys.append(head)
            samples.append(ac + ct + dpc + [0] * 6 + ld + mac + rsi)
            if verbose:
                inf = interface(hmmscan.records[0])
                infs.append(inf)
        finally:
            os.remove(temp)
        scaler = joblib.load(PATH + 'data/scaler.pkl')
    return keys, scaler.transform(samples), infs


def predict(fasta, db, verbose):
    keys, samples, infs = encode(fasta, db, verbose=verbose)
    model = joblib.load(PATH + 'data/model.pkl')
    scores = model.predict_log_proba(samples[:, rank - 1])[:, 1]
    scores = np.where(np.isfinite(scores), scores + 7, 0)
    return keys, scores, infs


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvm:i:l:o:n:d:")
        OPT = dict(opts)
    except getopt.GetoptError:
        print 'ERROR: Invalid arguments usage. Please type \'-h\' for help.'
        sys.exit(2)
    if '-h' in OPT:
        print USAGE + '\n' + VERSION
    else:
        if '-i' not in OPT:
            print 'ERROR: Input file is missing. Please type \'-h\' for help.'
            sys.exit(2)
        elif not os.path.exists(OPT['-i']):
            print 'ERROR: Input file cannot be found. Please type \'-h\' for help.'
            sys.exit(2)
        elif not os.path.exists(OPT['-d']):
            print 'ERROR: HMMER database cannot be found. Please type \'-h\' for help.'
            sys.exit(2)
        # Process train and predict submit
        else:
            verbose = True if '-v' in OPT else False
            keys, scores, infs = predict(OPT['-i'], OPT['-d'], verbose)
            results = ''
            for i, key in enumerate(keys):
                flag90 = '>=' if scores[i] >= CUTOFF90 else '<'
                flag95 = '>=' if scores[i] >= CUTOFF95 else '<'
                type = 'YES' if scores[i] >= CUTOFF90 else 'NO'
                if type == 'YES':
                    sig = ' ( SIGNIFICANT: YES )' if scores[i] >= CUTOFF95 else ' ( SIGNIFICANT: NO )'
                    type += sig
                results += '//\n'
                results += 'NAME       %s\n' % key
                results += 'SCORE      %f %s %f ( SPECIFICITY = 90%% ) && %s %f ( SPECIFICITY = 95%% )\n' % (
                            scores[i], flag90, CUTOFF90, flag95, CUTOFF95)
                results += 'TYPE       %s\n' % type
                if verbose:
                    results += infs[i]
            if '-o' not in OPT:
                print results
            else:
                output = open(OPT['-o'], 'w')
                output.write(results)
                output.close()


if __name__ == '__main__': main()
